@@ -79,10 +79,14 @@
   compatibility entries for SoundBlaster CD-ROM, netlink, SCSI generic and
   SCSI tape drivers.
 
-    Last updated by Richard Gooch   30-OCT-1999: Added entries for /dev/st,
+    Updated by      Richard Gooch   30-OCT-1999: Added entries for /dev/st,
   /dev/sg and /dev/sr hierarchies. Added compatibility entries for loop, IDE
   tape and SCSI CD-ROM drivers.
 
+    Last updated by Richard Gooch   31-OCT-1999: Added compatibility entries
+  for floppy, RAMDISC, meta-device, SCSI disc, IDE disc and IDE CD-ROM drivers.
+  Added entries for /dev/sd, /dev/ide/hd and /dev/ide/cd hierarchies.
+
 
 */
 #include <stdio.h>
@@ -101,6 +105,7 @@
 #include <fcntl.h>
 #include <linux/devfs_fs.h>
 #include <linux/kdev_t.h>
+#include <linux/major.h>
 #include <syslog.h>
 #include <signal.h>
 #include <regex.h>
@@ -108,7 +113,7 @@
 #include <karma.h>
 
 
-#define VERSION "1.2.7"
+#define VERSION "1.2.8"
 #define CONFIG_FILE "/etc/devfsd.conf"
 #define MAX_ARGS (6 + 1)
 
@@ -132,9 +137,8 @@
 #define MKACTION(what,when) (struct action_type) {what, when}
 
 #define SYSLOG(level, format, args...) \
-    {if (syslog_is_open) syslog (level, format, ##args); \
-     else fprintf (stderr, format, ## args); \
-    }
+    if (syslog_is_open) syslog (level, format, ##args); \
+    else fprintf (stderr, format, ## args)
 
 
 struct permissions_type
@@ -805,14 +809,14 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	}
 	else if (strncmp (info->devname, "netlink/", 8) == 0)
 	    compat_name = info->devname + 8;
-	else if ( (strncmp (info->devname, "scsi/", 4) == 0) &&
+	else if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
 		  (strcmp (info->devname + info->namelen - 7,"generic") == 0) )
 	{
 	    sprintf (compat_buf, "sg%u", info->minor);
 	    compat_name = compat_buf;
 	}
-	else if ( (strncmp (info->devname, "scsi/", 4) == 0) &&
-		  (strncmp (ptr = strrchr (info->devname, '/'), "mt", 2)
+	else if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
+		  (strncmp (ptr = (strrchr (info->devname, '/') + 1), "mt", 2)
 		   == 0) )
 	{
 	    char mode = ptr[2];
@@ -836,17 +840,181 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    compat_name = compat_buf;
 	    if (rewind != 'n') ++compat_name;
 	}
-	else if ( (strncmp (info->devname, "scsi/", 4) == 0) &&
+	else if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
 		  (strcmp (info->devname + info->namelen - 2, "cd") == 0) )
 	{
 	    sprintf (compat_buf, "sr%u", info->minor);
 	    compat_name = compat_buf;
 	}
+	else if (strncmp (info->devname, "floppy/", 7) == 0)
+	{
+	    sprintf (compat_buf, "fd%s", info->devname + 7);
+	    compat_name = compat_buf;
+	}
+	else if (strncmp (info->devname, "rd/", 3) == 0)
+	{
+	    sprintf (compat_buf, "ram%s", info->devname + 3);
+	    compat_name = compat_buf;
+	}
+	else if (strncmp (info->devname, "md/", 3) == 0)
+	{
+	    sprintf (compat_buf, "md%s", info->devname + 3);
+	    compat_name = compat_buf;
+	}
+	else if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
+		  (strcmp (info->devname + info->namelen - 4, "disc") == 0) )
+	{
+	    sprintf ( compat_buf, "sd%c", 'a' + (info->minor >> 4) );
+	    compat_name = compat_buf;
+	}
+	else if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
+		  (strncmp (ptr = (strrchr (info->devname, '/') +1), "part", 4)
+		   == 0) )
+	{
+	    sprintf (compat_buf, "sd%c%s", 'a' + (info->minor >> 4), ptr + 4);
+	    compat_name = compat_buf;
+	}
+	else if ( (strncmp (info->devname, "ide/", 4) == 0) &&
+		  (strcmp (info->devname + info->namelen - 4, "disc") == 0) )
+	{
+	    char letter;
+
+	    switch (info->major)
+	    {
+	      case IDE0_MAJOR:
+		letter = 'a';
+		break;
+	      case IDE1_MAJOR:
+		letter = 'c';
+		break;
+	      case IDE2_MAJOR:
+		letter = 'e';
+		break;
+	      case IDE3_MAJOR:
+		letter = 'g';
+		break;
+	      case IDE4_MAJOR:
+		letter = 'i';
+		break;
+	      case IDE5_MAJOR:
+		letter = 'k';
+		break;
+	      case IDE6_MAJOR:
+		letter = 'm';
+		break;
+	      case IDE7_MAJOR:
+		letter = 'o';
+		break;
+	      case IDE8_MAJOR:
+		letter = 'q';
+		break;
+	      case IDE9_MAJOR:
+		letter = 's';
+		break;
+	      default:
+		letter = 'y';
+		break;
+	    }
+	    if (info->minor > 63) ++letter;
+	    sprintf (compat_buf, "hd%c", letter);
+	    compat_name = compat_buf;
+	}
+	else if ( (strncmp (info->devname, "ide/", 4) == 0) &&
+		  (strncmp (ptr = (strrchr (info->devname, '/') +1), "part", 4)
+		   == 0) )
+	{
+	    char letter;
+
+	    switch (info->major)
+	    {
+	      case IDE0_MAJOR:
+		letter = 'a';
+		break;
+	      case IDE1_MAJOR:
+		letter = 'c';
+		break;
+	      case IDE2_MAJOR:
+		letter = 'e';
+		break;
+	      case IDE3_MAJOR:
+		letter = 'g';
+		break;
+	      case IDE4_MAJOR:
+		letter = 'i';
+		break;
+	      case IDE5_MAJOR:
+		letter = 'k';
+		break;
+	      case IDE6_MAJOR:
+		letter = 'm';
+		break;
+	      case IDE7_MAJOR:
+		letter = 'o';
+		break;
+	      case IDE8_MAJOR:
+		letter = 'q';
+		break;
+	      case IDE9_MAJOR:
+		letter = 's';
+		break;
+	      default:
+		letter = 'y';
+		break;
+	    }
+	    if (info->minor > 63) ++letter;
+	    sprintf (compat_buf, "hd%c%s", letter, ptr + 4);
+	    compat_name = compat_buf;
+	}
+	else if ( (strncmp (info->devname, "ide/", 4) == 0) &&
+		  (strcmp (info->devname + info->namelen - 2, "cd") == 0) )
+	{
+	    char letter;
+
+	    switch (info->major)
+	    {
+	      case IDE0_MAJOR:
+		letter = 'a';
+		break;
+	      case IDE1_MAJOR:
+		letter = 'c';
+		break;
+	      case IDE2_MAJOR:
+		letter = 'e';
+		break;
+	      case IDE3_MAJOR:
+		letter = 'g';
+		break;
+	      case IDE4_MAJOR:
+		letter = 'i';
+		break;
+	      case IDE5_MAJOR:
+		letter = 'k';
+		break;
+	      case IDE6_MAJOR:
+		letter = 'm';
+		break;
+	      case IDE7_MAJOR:
+		letter = 'o';
+		break;
+	      case IDE8_MAJOR:
+		letter = 'q';
+		break;
+	      case IDE9_MAJOR:
+		letter = 's';
+		break;
+	      default:
+		letter = 'y';
+		break;
+	    }
+	    if (info->minor > 63) ++letter;
+	    sprintf (compat_buf, "hd%c", letter);
+	    compat_name = compat_buf;
+	}
 	break;
       case AC_MKNEWCOMPAT:
       case AC_RMNEWCOMPAT:
-	if ( (strncmp (info->devname, "scsi/", 4) == 0) &&
-	     (strncmp (ptr = strrchr (info->devname, '/'), "mt", 2)
+	if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
+	     (strncmp (ptr = (strrchr (info->devname, '/') + 1), "mt", 2)
 	      == 0) )
 	{
 	    int mode, host, bus, target, lun;
@@ -876,7 +1044,7 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    sprintf (dest_buf, "../%s", info->devname);
 	    dest_name = dest_buf;
 	}
-	else if ( (strncmp (info->devname, "scsi/", 4) == 0) &&
+	else if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
 		  (strcmp (info->devname + info->namelen - 7,"generic") == 0) )
 	{
 	    int host, bus, target, lun;
@@ -888,7 +1056,7 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    sprintf (dest_buf, "../%s", info->devname);
 	    dest_name = dest_buf;
 	}
-	else if ( (strncmp (info->devname, "scsi/", 4) == 0) &&
+	else if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
 		  (strcmp (info->devname + info->namelen - 2, "cd") == 0) )
 	{
 	    int host, bus, target, lun;
@@ -900,6 +1068,78 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    sprintf (dest_buf, "../%s", info->devname);
 	    dest_name = dest_buf;
 	}
+	else if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
+		  (strcmp (info->devname + info->namelen - 4, "disc") == 0) )
+	{
+	    int host, bus, target, lun;
+
+	    sscanf (info->devname + 5, "host%d/bus%d/target%d/lun%d/",
+		    &host, &bus, &target, &lun);
+	    sprintf (compat_buf, "sd/c%db%dt%du%d", host, bus, target, lun);
+	    compat_name = compat_buf;
+	    sprintf (dest_buf, "../%s", info->devname);
+	    dest_name = dest_buf;
+	}
+	else if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
+		  (strncmp (ptr = (strrchr (info->devname, '/') +1), "part", 4)
+		   == 0) )
+	{
+	    int host, bus, target, lun, partition;
+
+	    sscanf (info->devname + 5, "host%d/bus%d/target%d/lun%d/part%d",
+		    &host, &bus, &target, &lun, &partition);
+	    sprintf (compat_buf, "sd/c%db%dt%du%dp%d",
+		     host, bus, target, lun, partition);
+	    compat_name = compat_buf;
+	    sprintf (dest_buf, "../%s", info->devname);
+	    dest_name = dest_buf;
+	}
+
+	else if ( (strncmp (info->devname, "ide/", 4) == 0) &&
+		  (strcmp (info->devname + info->namelen - 4, "disc") == 0) )
+	{
+	    int host, bus, target, lun;
+
+	    sscanf (info->devname + 4, "host%d/bus%d/target%d/lun%d/",
+		    &host, &bus, &target, &lun);
+	    sprintf (compat_buf, "%s/ide/hd", mount_point);
+	    mkdir (compat_buf,S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
+	    sprintf (compat_buf, "ide/hd/c%db%dt%du%d",
+		     host, bus, target, lun);
+	    compat_name = compat_buf;
+	    sprintf (dest_buf, "../%s", info->devname + 4);
+	    dest_name = dest_buf;
+	}
+	else if ( (strncmp (info->devname, "ide/", 4) == 0) &&
+		  (strncmp (ptr = (strrchr (info->devname, '/') +1), "part", 4)
+		   == 0) )
+	{
+	    int host, bus, target, lun, partition;
+
+	    sscanf (info->devname + 4, "host%d/bus%d/target%d/lun%d/part%d",
+		    &host, &bus, &target, &lun, &partition);
+	    sprintf (compat_buf, "%s/ide/hd", mount_point);
+	    mkdir (compat_buf,S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
+	    sprintf (compat_buf, "ide/hd/c%db%dt%du%dp%d",
+		     host, bus, target, lun, partition);
+	    compat_name = compat_buf;
+	    sprintf (dest_buf, "../%s", info->devname + 4);
+	    dest_name = dest_buf;
+	}
+	else if ( (strncmp (info->devname, "ide/", 4) == 0) &&
+		  (strcmp (info->devname + info->namelen - 2, "cd") == 0) )
+	{
+	    int host, bus, target, lun;
+
+	    sscanf (info->devname + 4, "host%d/bus%d/target%d/lun%d/",
+		    &host, &bus, &target, &lun);
+	    sprintf (compat_buf, "%s/ide/cd", mount_point);
+	    mkdir (compat_buf,S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
+	    sprintf (compat_buf, "ide/cd/c%db%dt%du%d", host, bus, target,lun);
+	    compat_name = compat_buf;
+	    sprintf (dest_buf, "../%s", info->devname + 4);
+	    dest_name = dest_buf;
+	}
 	break;
     }
     if (compat_name == NULL) return;
@@ -911,8 +1151,12 @@ static void action_compat (const struct devfsd_notify_struct *info,
       case AC_MKNEWCOMPAT:
 	if (symlink (dest_name, compat_path) != 0)
 	{
-	    SYSLOG (LOG_ERR, "%s: error creating symlink: \"%s\"\t%s\n",
-		    function_name, compat_path, ERRSTRING);
+	    if (errno == EEXIST)
+		SYSLOG (LOG_INFO, "symlink: \"%s\" already exists\n",
+			compat_path);
+	    else
+		SYSLOG (LOG_ERR, "%s: error creating symlink: \"%s\"\t%s\n",
+			function_name, compat_path, ERRSTRING);
 	}
 	else SYSLOG (LOG_INFO, "made symlink: \"%s\" for dev: %u,%u\n",
 		     compat_path, info->major, info->minor);
@@ -927,7 +1171,6 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	else SYSLOG (LOG_INFO, "unlinked: \"%s\"\n", compat_path);
 	break;
     }
-    
 }   /*  End Function action_compat  */
 
 static void free_config ()
